import logging
from typing import TYPE_CHECKING, Iterator, List, Optional

from pyminion.core import Card
from pyminion.exceptions import CardNotFound, EmptyPile
from pyminion.players import Player

if TYPE_CHECKING:
    from pyminion.game import Game

logger = logging.getLogger()


class Bot(Player):
    """
    Abstract representation of Bot.

    Defines at a high level how a bot can play and buy cards.
    The intention is to inherit from this class to make concrete bot implementations.

    Does not implement any logic for playing or buying cards, as those functions
    (action_priority and buy_priority), are meant to be implemented when defining new bots.

    Does implement default responses to any decision the bot might have to make
    when playing a card or responding to an attack as to not crash the game.

    """

    def __init__(
        self,
        player_id: str = "bot",
    ):
        super().__init__(player_id=player_id)

    def action_priority(self, game: "Game") -> Iterator[Card]:
        """
        Add logic for playing action cards through this method

        This function should be a generator where each call
        yields a desired card to play if conditions are met

        """
        raise NotImplementedError

    def start_action_phase(self, game: "Game"):
        """
        Attempts to play cards from the action_priority queue if possible

        """
        viable_actions = [card for card in self.hand.cards if "Action" in card.type]
        logger.info(f"{self.player_id}'s hand: {self.hand}")
        while viable_actions and self.state.actions:
            for card in self.action_priority(game=game):
                try:
                    self.play(target_card=card, game=game)
                except CardNotFound:
                    pass
                if not self.state.actions:
                    return
            return

    def start_treasure_phase(self, game: "Game"):
        """
        At start of action phase, bot simply plays all of their treasures

        """
        viable_treasures = [card for card in self.hand.cards if "Treasure" in card.type]
        self.autoplay_treasures(viable_treasures=viable_treasures, game=game)

    def buy_priority(self, game: "Game") -> Iterator[Card]:
        """
        Add logic for buy priority through this method

        This function should be a generator where each call
        yields a desired card to buy if conditions are met

        """
        raise NotImplementedError

    def start_buy_phase(self, game: "Game"):
        """
        Attempts to buy cards from the buy_priority queue if possible

        """
        logger.info(f"{self.player_id} has {self.state.money} money")

        while self.state.buys:
            for card in self.buy_priority(game=game):
                try:
                    self.buy(card, supply=game.supply)
                    break

                except EmptyPile:
                    pass

            else:
                logger.info(f"{self} buys nothing")
                return

    def take_turn(self, game: "Game") -> None:
        self.start_turn()
        logger.info(f"\nTurn {self.turns} - {self.player_id}")
        self.start_action_phase(game)
        self.start_treasure_phase(game)
        self.start_buy_phase(game)
        self.start_cleanup_phase()

    # DEFAULT RESPONSES
    # These methods can be implemented with specific game logic
    # when creating new bots. In this class, these methods just return
    # a valid response as to not crash the game.

    def binary_resp(
        self, game: "Game", card: Card, relevant_cards: Optional[List[Card]] = None
    ) -> bool:
        return True

    def discard_resp(
        self,
        card: Card,
        valid_cards: List[Card],
        game: "Game",
        required: bool = True,
    ) -> Optional[Card]:
        if required:
            return valid_cards[0]
        else:
            return None

    def multiple_discard_resp(
        self,
        card: Card,
        valid_cards: List[Card],
        game: "Game",
        num_discard: Optional[int] = None,
        required: bool = True,
    ) -> Optional[List[Card]]:
        if required:
            return valid_cards[:num_discard]
        else:
            return None

    def gain_resp(
        self,
        card: Card,
        valid_cards: List[Card],
        game: "Game",
        required: bool = True,
    ) -> Optional[Card]:
        if required:
            return valid_cards[0]
        else:
            return None

    def multiple_gain_resp(
        self,
        card: Card,
        valid_cards: List[Card],
        game: "Game",
        num_gain: Optional[int] = None,
        required: bool = True,
    ) -> Optional[List[Card]]:
        if required:
            return valid_cards[:num_gain]
        else:
            return None

    def trash_resp(
        self,
        card: Card,
        valid_cards: List[Card],
        game: "Game",
        required: bool = True,
    ) -> Optional[Card]:
        if required:
            return valid_cards[0]
        else:
            return None

    def multiple_trash_resp(
        self,
        card: Card,
        valid_cards: List[Card],
        game: "Game",
        num_trash: Optional[int] = None,
        required: bool = True,
    ) -> Optional[List[Card]]:
        if required:
            return valid_cards[:num_trash]
        else:
            return None

    def topdeck_resp(
        self,
        card: Card,
        valid_cards: List[Card],
        game: "Game",
        required: bool = True,
    ) -> Optional[Card]:
        if required:
            return valid_cards[0]
        else:
            return None

    def double_play_resp(
        self,
        card: Card,
        valid_cards: List[Card],
        game: "Game",
        required: bool = True,
    ) -> Optional[Card]:
        if required:
            return valid_cards[0]
        else:
            return None

    def is_attacked(self, player: Player, attack_card: Card) -> bool:
        return True



import logging
import math
from typing import TYPE_CHECKING, List, Optional, Tuple, Union

from pyminion.bots.bot import Bot
from pyminion.core import AbstractDeck, Action, Card, Treasure, Victory
from pyminion.decisions import validate_input
from pyminion.exceptions import (EmptyPile, InvalidBotImplementation,
                                 InvalidMultiCardInput, InvalidSingleCardInput)
from pyminion.players import Human, Player

if TYPE_CHECKING:
    from pyminion.game import Game


logger = logging.getLogger()


class Copper(Treasure):
    def __init__(
        self,
        name: str = "Copper",
        cost: int = 0,
        type: Tuple[str] = ("Treasure",),
        money: int = 1,
    ):
        super().__init__(name, cost, type, money)

    def play(self, player: Player, game: "Game"):
        player.playmat.add(self)
        player.hand.remove(self)
        player.state.money += self.money


class Silver(Treasure):
    def __init__(
        self,
        name: str = "Silver",
        cost: int = 3,
        type: Tuple[str] = ("Treasure",),
        money: int = 2,
    ):
        super().__init__(name, cost, type, money)

    def play(self, player: Player, game: "Game"):

        # check if this is the first silver played and if there are any merchants in play
        if self not in player.playmat.cards:
            if num_merchants := len(
                [card for card in player.playmat.cards if card.name == "Merchant"]
            ):
                player.state.money += num_merchants

        player.playmat.add(self)
        player.hand.remove(self)
        player.state.money += self.money


class Gold(Treasure):
    def __init__(
        self,
        name: str = "Gold",
        cost: int = 6,
        type: Tuple[str] = ("Treasure",),
        money: int = 3,
    ):
        super().__init__(name, cost, type, money)

    def play(self, player: Player, game: "Game"):
        player.playmat.add(self)
        player.hand.remove(self)
        player.state.money += self.money


class Estate(Victory):
    def __init__(
        self,
        name: str = "Estate",
        cost: int = 2,
        type: Tuple[str] = ("Victory",),
    ):
        super().__init__(name, cost, type)

    def score(self, player: Player) -> int:
        vp = 1
        return vp


class Duchy(Victory):
    def __init__(
        self,
        name: str = "Duchy",
        cost: int = 5,
        type: Tuple[str] = ("Victory",),
    ):
        super().__init__(name, cost, type)

    def score(self, player: Player) -> int:
        vp = 3
        return vp


class Province(Victory):
    def __init__(
        self,
        name: str = "Province",
        cost: int = 8,
        type: Tuple[str] = ("Victory",),
    ):
        super().__init__(name, cost, type)

    def score(self, player: Player) -> int:
        vp = 6
        return vp


class Curse(Victory):
    def __init__(
        self,
        name: str = "Curse",
        cost: int = 0,
        type: Tuple[str] = ("Curse",),
    ):
        super().__init__(name, cost, type)

    def score(self, player: Player) -> int:
        vp = -1
        return vp


class Gardens(Victory):
    """
    Worth 1VP for every 10 cards you have (round down)

    """

    def __init__(
        self,
        name: str = "Gardens",
        cost: int = 4,
        type: Tuple[str] = ("Victory",),
    ):
        super().__init__(name, cost, type)

    def score(self, player: Player) -> int:
        total_count = len(player.get_all_cards())
        vp = math.floor(total_count / 10)
        return vp


class Smithy(Action):
    """
    + 3 cards

    """

    def __init__(
        self,
        name: str = "Smithy",
        cost: int = 4,
        type: Tuple[str] = ("Action",),
        draw: int = 3,
    ):
        super().__init__(name, cost, type, draw=draw)

    def play(
        self, player: Union[Human, Bot], game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.draw(3)


class Village(Action):
    """
    + 1 card, + 2 actions

    """

    def __init__(
        self,
        name: str = "Village",
        cost: int = 3,
        type: Tuple[str] = ("Action",),
        actions: int = 2,
        draw: int = 1,
    ):
        super().__init__(name, cost, type, actions=actions, draw=draw)

    def play(
        self, player: Union[Human, Bot], game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.state.actions += 2
        player.draw()


class Laboratory(Action):
    """
    +2 cards, +1 action

    """

    def __init__(
        self,
        name: str = "Laboratory",
        cost: int = 5,
        type: Tuple[str] = ("Action",),
        actions: int = 1,
        draw: int = 2,
    ):
        super().__init__(name, cost, type, actions=actions, draw=draw)

    def play(
        self, player: Union[Human, Bot], game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.state.actions += 1
        player.draw(2)


class Market(Action):
    """
    +1 card, +1 action, +1 money, +1 buy

    """

    def __init__(
        self,
        name: str = "Market",
        cost: int = 5,
        type: Tuple[str] = ("Action",),
        actions: int = 1,
        draw: int = 1,
        money: int = 1,
    ):
        super().__init__(name, cost, type, actions, draw, money)

    def play(
        self, player: Union[Human, Bot], game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.state.actions += 1
        player.draw()
        player.state.money += 1
        player.state.buys += 1


class Moneylender(Action):
    """
    You may trash a copper from your hand for + 3 money

    """

    def __init__(
        self,
        name: str = "Moneylender",
        cost: int = 4,
        type: Tuple[str] = ("Action",),
    ):
        super().__init__(name, cost, type)

    def play(
        self,
        player: Union[Human, Bot],
        game: "Game",
        generic_play: bool = True,
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        if copper not in player.hand.cards:
            return

        if isinstance(player, Human):
            response = player.binary_decision(
                prompt="Do you want to trash a copper from your hand for +3 money? y/n: ",
            )

        elif isinstance(player, Bot):
            response = player.binary_resp(game=game, card=self)

        if response:
            player.trash(target_card=copper, trash=game.trash)
            player.state.money += 3


class Cellar(Action):
    """
    +1 Action

    Discard any number of cards, then draw that many

    """

    def __init__(
        self,
        name: str = "Cellar",
        cost: int = 2,
        type: Tuple[str] = ("Action",),
        actions: int = 1,
    ):
        super().__init__(name, cost, type, actions=actions)

    def play(
        self,
        player: Union[Human, Bot],
        game: "Game",
        generic_play: bool = True,
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.state.actions += 1

        if not player.hand.cards:
            return

        if isinstance(player, Human):
            discard_cards = player.multiple_card_decision(
                prompt="Enter the cards you would like to discard seperated by commas: ",
                valid_cards=player.hand.cards,
            )

        elif isinstance(player, Bot):
            discard_cards = player.multiple_discard_resp(
                card=self,
                valid_cards=player.hand.cards,
                game=game,
                required=False,
            )

        if discard_cards:
            for card in discard_cards:
                player.discard(card)
            player.draw(len(discard_cards))


class Chapel(Action):
    """
    Trash up to 4 cards from your hand

    """

    def __init__(
        self,
        name: str = "Chapel",
        cost: int = 2,
        type: Tuple[str] = ("Action",),
    ):
        super().__init__(name, cost, type)

    def play(
        self, player: Union[Human, Bot], game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")
        if generic_play:
            super().generic_play(player)

        if not player.hand.cards:
            return

        @validate_input(exceptions=InvalidMultiCardInput)
        def get_trash_cards() -> Optional[List[Card]]:

            trash_cards = player.multiple_card_decision(
                prompt="Enter up to 4 cards you would like to trash from your hand: ",
                valid_cards=player.hand.cards,
            )

            if trash_cards and len(trash_cards) > 4:
                raise InvalidMultiCardInput(
                    "You cannot trash more than 4 cards with chapel"
                )

            return trash_cards

        if isinstance(player, Human):
            trash_cards = get_trash_cards()

        elif isinstance(player, Bot):
            trash_cards = player.multiple_trash_resp(
                card=self,
                valid_cards=player.hand.cards,
                game=game,
                required=False,
            )
            if trash_cards and len(trash_cards) > 4:
                raise InvalidMultiCardInput(
                    "Attempted to trash more than 4 cards with chapel"
                )
        if not trash_cards:
            return
        for card in trash_cards:
            player.trash(card, game.trash)


class Workshop(Action):
    """
    Gain a card costing up to 4 money

    """

    def __init__(
        self,
        name: str = "Workshop",
        cost: int = 3,
        type: Tuple[str] = ("Action",),
    ):
        super().__init__(name, cost, type)

    def play(self, player: Player, game: "Game", generic_play: bool = True) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        @validate_input(exceptions=InvalidSingleCardInput)
        def get_gain_card() -> Card:

            gain_card = player.single_card_decision(
                prompt="Gain a card costing up to 4 money: ",
                valid_cards=game.supply.avaliable_cards(),
            )

            if not gain_card:
                raise InvalidSingleCardInput("You must gain a card")
            if gain_card.cost > 4:
                raise InvalidSingleCardInput("Card must cost less than 4 money")

            return gain_card

        if isinstance(player, Human):
            gain_card = get_gain_card()

        elif isinstance(player, Bot):
            gain_card = player.gain_resp(
                card=self,
                valid_cards=[
                    card for card in game.supply.avaliable_cards() if card.cost <= 4
                ],
                game=game,
                required=True,
            )
            if not gain_card:
                raise InvalidSingleCardInput("You must gain a card")
            if gain_card.cost > 4:
                raise InvalidSingleCardInput("Card must cost less than 4 money")

        player.gain(gain_card, game.supply)


class Festival(Action):
    """
    + 2 actions, + 1 buy, + 2 money

    """

    def __init__(
        self,
        name: str = "Festival",
        cost: int = 5,
        type: Tuple[str] = ("Action",),
        actions: int = 2,
        money: int = 2,
    ):
        super().__init__(name, cost, type, actions=actions, money=money)

    def play(
        self, player: Union[Human, Bot], game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.state.actions += 2
        player.state.money += 2
        player.state.buys += 1


class Harbinger(Action):
    """
    +1 card, +1 action

    Look through your discard pile. You may put a card from it onto your deck

    """

    def __init__(
        self,
        name: str = "Harbinger",
        cost: int = 3,
        type: Tuple[str] = ("Action",),
        actions: int = 1,
        draw: int = 1,
    ):
        super().__init__(name, cost, type, actions=actions, draw=draw)

    def play(
        self, player: Union[Human, Bot], game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.state.actions += 1
        player.draw()

        if not player.discard_pile:
            return

        if isinstance(player, Human):
            topdeck_card = player.single_card_decision(
                prompt="You may select a card from your discard pile to put onto your deck: ",
                valid_cards=player.discard_pile.cards,
            )

        elif isinstance(player, Bot):
            topdeck_card = player.topdeck_resp(
                card=self,
                valid_cards=player.discard_pile.cards,
                game=game,
                required=False,
            )

        if not topdeck_card or isinstance(topdeck_card, str):
            return

        player.deck.add(player.discard_pile.remove(topdeck_card))


class Vassal(Action):
    """
    +2 money

    Discard the top card of your deck. If it's an action card you may play it.

    """

    def __init__(
        self,
        name: str = "Vassal",
        cost: int = 3,
        type: Tuple[str] = ("Action",),
        money: int = 2,
    ):
        super().__init__(name, cost, type, money=money)

    def play(
        self, player: Union[Human, Bot], game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.state.money += 2
        player.draw(destination=player.discard_pile, silent=True)

        if not player.discard_pile:
            return

        discard_card = player.discard_pile.cards[-1]

        logger.info(f"{player} discards {discard_card}")

        if "Action" not in discard_card.type:
            return

        if isinstance(player, Human):
            decision = player.binary_decision(
                prompt=f"You discarded {discard_card.name}, would you like to play it? (y/n):  "
            )
        elif isinstance(player, Bot):
            decision = player.binary_resp(game=game, card=self)

        if not decision:
            return

        played_card = player.discard_pile.cards.pop()
        player.playmat.add(played_card)
        player.exact_play(card=player.playmat.cards[-1], game=game, generic_play=False)
        return


class Artisan(Action):
    """
    Gain a card to your hand costing up to 5 money.

    Put a card from your hand onto your deck

    """

    def __init__(
        self,
        name: str = "Artisan",
        cost: int = 6,
        type: Tuple[str] = ("Action",),
    ):
        super().__init__(name, cost, type)

    def play(
        self, player: Union[Human, Bot], game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        @validate_input(exceptions=InvalidSingleCardInput)
        def get_gain_card() -> Card:
            gain_card = player.single_card_decision(
                prompt="Gain a card costing up to 5 money: ",
                valid_cards=game.supply.avaliable_cards(),
            )
            if not gain_card or isinstance(gain_card, str):
                raise InvalidSingleCardInput("You must gain a card")
            if gain_card.cost > 4:
                raise InvalidSingleCardInput("Card must cost at most 5 money")
            return gain_card

        @validate_input(exceptions=InvalidSingleCardInput)
        def get_topdeck_card() -> Card:
            topdeck_card = player.single_card_decision(
                prompt="Put a card from your hand onto your deck: ",
                valid_cards=player.hand.cards,
            )
            if not topdeck_card or isinstance(topdeck_card, str):
                raise InvalidSingleCardInput("You must topdeck a card")
            return topdeck_card

        if isinstance(player, Human):
            gain_card = get_gain_card()
            player.gain(card=gain_card, supply=game.supply, destination=player.hand)
            topdeck_card = get_topdeck_card()

        elif isinstance(player, Bot):
            gain_card = player.gain_resp(
                card=self,
                valid_cards=[
                    card for card in game.supply.avaliable_cards() if card.cost <= 5
                ],
                game=game,
                required=True,
            )

            if not gain_card:
                raise InvalidSingleCardInput("Must gain a card with Artisan")
            if gain_card.cost > 5:
                raise InvalidSingleCardInput("Card must cost at most 5 money")

            player.gain(card=gain_card, supply=game.supply, destination=player.hand)

            topdeck_card = player.topdeck_resp(
                card=self,
                valid_cards=player.hand.cards,
                game=game,
                required=True,
            )

        for card in player.hand.cards:
            if card == topdeck_card:
                player.deck.add(player.hand.remove(card))
                return


class Poacher(Action):
    """
    +1 card, +1 action, + 1 money

    Discard a card per empty Supply pile

    """

    def __init__(
        self,
        name: str = "Poacher",
        cost: int = 4,
        type: Tuple[str] = ("Action",),
        actions: int = 1,
        draw: int = 1,
        money: int = 1,
    ):
        super().__init__(name, cost, type, actions, draw, money)

    def play(
        self, player: Union[Human, Bot], game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.draw()
        player.state.actions += 1
        player.state.money += 1

        empty_piles = game.supply.num_empty_piles()

        if empty_piles == 0:
            return

        discard_num = min(empty_piles, len(player.hand))

        if isinstance(player, Human):

            @validate_input(exceptions=InvalidMultiCardInput)
            def get_discard_cards() -> List[Card]:
                discard_cards = player.multiple_card_decision(
                    prompt=f"Discard {discard_num} card(s) from your hand: ",
                    valid_cards=player.hand.cards,
                )
                if not discard_cards or len(discard_cards) != discard_num:
                    raise InvalidMultiCardInput(
                        f"You must discard {discard_num} card(s)"
                    )

                return discard_cards

            discard_cards = get_discard_cards()

        elif isinstance(player, Bot):
            discard_cards = player.multiple_discard_resp(
                card=self,
                valid_cards=player.hand.cards,
                game=game,
                num_discard=discard_num,
                required=True,
            )
            if not discard_cards or len(discard_cards) != discard_num:
                raise InvalidMultiCardInput(f"You must discard {discard_num} card(s)")

        for discard_card in discard_cards:
            player.discard(discard_card)


class CouncilRoom(Action):
    """
    +4 cards, +1 buy

    Each other player draws a card

    """

    def __init__(
        self,
        name: str = "Council Room",
        cost: int = 5,
        type: Tuple[str] = ("Action",),
        draw: int = 4,
    ):
        super().__init__(name, cost, type, draw=draw)

    def play(
        self, player: Union[Human, Bot], game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.draw(4)
        player.state.buys += 1

        for p in game.players:
            if p is not player:
                p.draw()


class Witch(Action):
    """
    +2 cards

    Each other player gains a curse

    """

    def __init__(
        self,
        name: str = "Witch",
        cost: int = 5,
        type: Tuple[str] = ("Action", "Attack"),
        draw: int = 2,
    ):
        super().__init__(name, cost, type, draw=draw)

    def play(
        self, player: Union[Human, Bot], game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.draw(2)

        for opponent in game.players:
            if opponent is not player:
                if opponent.is_attacked(player=player, attack_card=self):

                    # attempt to gain a curse. if curse pile is empty, proceed
                    try:
                        opponent.gain(
                            card=curse,
                            supply=game.supply,
                        )
                    except EmptyPile:
                        pass


class Moat(Action):
    """
    +2 cards

    When another player plays an attack card, you may first
    reveal this from your hand, to be unaffected by it

    """

    def __init__(
        self,
        name: str = "Moat",
        cost: int = 2,
        type: Tuple[str] = ("Action", "Reaction"),
        draw: int = 2,
    ):
        super().__init__(name, cost, type, draw=draw)

    def play(
        self, player: Union[Human, Bot], game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.draw(2)


class Merchant(Action):
    """
    +1 card, +1 action

    The first time you play a Silver this turn, +1 money

    """

    def __init__(
        self,
        name: str = "Merchant",
        cost: int = 3,
        type: Tuple[str] = ("Action",),
        actions: int = 1,
        draw: int = 1,
    ):
        super().__init__(name, cost, type, actions=actions, draw=draw)

    def play(
        self, player: Union[Human, Bot], game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.draw(1)
        player.state.actions += 1


class Bandit(Action):
    """
    Gain a Gold. Each other player reveals the top 2 cards of their deck,
    trashes a revealed treasure other than Copper, and discards the rest

    """

    def __init__(
        self,
        name: str = "Bandit",
        cost: int = 5,
        type: Tuple[str] = ("Action", "Attack"),
    ):
        super().__init__(name, cost, type)

    def play(
        self, player: Union[Human, Bot], game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        # attempt to gain a gold. if gold pile is empty, proceed
        try:
            player.gain(card=gold, supply=game.supply)
        except EmptyPile:
            pass

        for opponent in game.players:
            if opponent is not player:
                if opponent.is_attacked(player=player, attack_card=self):

                    revealed_cards = AbstractDeck()
                    opponent.draw(num_cards=2, destination=revealed_cards, silent=True)

                    logger.info(f"{opponent} reveals {revealed_cards}")

                    trash_card = None
                    for card in revealed_cards.cards:
                        if card.name == "Silver":
                            trash_card = card
                        elif card.name == "Gold" and not trash_card:
                            trash_card = card
                        elif (
                            "Treasure" in card.type
                            and card.name != "Copper"
                            and not trash_card
                        ):
                            trash_card = card

                    if trash_card:
                        game.trash.add(revealed_cards.remove(trash_card))

                    revealed_cards.move_to(opponent.discard_pile)
                    del revealed_cards


class Bureaucrat(Action):
    """
    Gain a Silver onto your deck. Each other player reveals a victory card from
    their hand and puts it onto their deck (or reveals a hand with no victory cards)

    """

    def __init__(
        self,
        name: str = "Bureaucrat",
        cost: int = 4,
        type: Tuple[str] = ("Action", "Attack"),
    ):
        super().__init__(name, cost, type)

    def play(
        self, player: Union[Human, Bot], game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        # attempt to gain a silver. if silver pile is empty, proceed
        try:
            player.gain(card=silver, supply=game.supply, destination=player.deck)
        except EmptyPile:
            pass

        for opponent in game.players:
            if opponent is not player and opponent.is_attacked(
                player=player, attack_card=self
            ):

                victory_cards = []
                for card in opponent.hand.cards:
                    if "Victory" in card.type:
                        victory_cards.append(card)

                if not victory_cards:
                    # Log opponent revealed hand
                    return

                @validate_input(exceptions=InvalidSingleCardInput)
                def get_topdeck_card(opponent: Human) -> Card:
                    topdeck_card = opponent.single_card_decision(
                        prompt="You must topdeck a Victory card from your hand: ",
                        valid_cards=victory_cards,
                    )
                    if not topdeck_card or isinstance(topdeck_card, str):
                        raise InvalidSingleCardInput("You must topdeck a Victory card")

                    return topdeck_card

                if isinstance(opponent, Human):
                    topdeck_card = get_topdeck_card(opponent)

                elif isinstance(opponent, Bot):
                    topdeck_card = opponent.topdeck_resp(
                        card=self,
                        valid_cards=victory_cards,
                        game=game,
                        required=True,
                    )
                    if not topdeck_card:
                        raise InvalidSingleCardInput("You must topdeck a Victory card")

                opponent.deck.add(opponent.hand.remove(topdeck_card))


class ThroneRoom(Action):
    """
    You may play an Action card from your hand twice

    """

    def __init__(
        self,
        name: str = "Throne Room",
        cost: int = 4,
        type: Tuple[str] = ("Action",),
    ):
        super().__init__(name, cost, type)

    def play(
        self, player: Union[Human, Bot], game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        action_cards = [card for card in player.hand.cards if "Action" in card.type]

        if not action_cards:
            return

        if isinstance(player, Human):
            dp_card = player.single_card_decision(
                prompt="You may play an action card from your hand twice: ",
                valid_cards=action_cards,
            )

        elif isinstance(player, Bot):
            dp_card = player.double_play_resp(
                card=self,
                valid_cards=action_cards,
                game=game,
                required=True,
            )

        if not dp_card or isinstance(dp_card, str):
            return

        for card in player.hand.cards:
            if card.name == dp_card.name:
                player.playmat.add(player.hand.remove(card))
                for i in range(2):
                    player.exact_play(card=card, game=game, generic_play=False)
                return


class Remodel(Action):
    """
    Trash a card from your hand. Gain a card costing up to $2 more than it

    """

    def __init__(
        self,
        name: str = "Remodel",
        cost: int = 4,
        type: Tuple[str] = ("Action",),
    ):
        super().__init__(name, cost, type)

    def play(
        self, player: Union[Human, Bot], game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        if isinstance(player, Human):

            @validate_input(exceptions=InvalidSingleCardInput)
            def get_trash_card() -> Card:

                trash_card = player.single_card_decision(
                    prompt="Trash a card form your hand: ",
                    valid_cards=player.hand.cards,
                )

                if not trash_card or isinstance(trash_card, str):
                    raise InvalidSingleCardInput("You must trash a card")
                return trash_card

            @validate_input(exceptions=InvalidSingleCardInput)
            def get_gain_card(trash_card: Card) -> Card:

                gain_card = player.single_card_decision(
                    prompt=f"Gain a card costing up to {trash_card.cost + 2} money: ",
                    valid_cards=game.supply.avaliable_cards(),
                )

                if not gain_card or isinstance(gain_card, str):
                    raise InvalidSingleCardInput("You must gain a card")
                if gain_card.cost > trash_card.cost + 2:
                    raise InvalidSingleCardInput(
                        f"Card must cost less than {trash_card.cost + 2} money"
                    )
                return gain_card

            trash_card = get_trash_card()
            gain_card = get_gain_card(trash_card)

        elif isinstance(player, Bot):
            trash_card = player.trash_resp(
                card=self,
                valid_cards=player.hand.cards,
                game=game,
                required=True,
            )
            if not trash_card:
                raise InvalidBotImplementation(
                    "Card must be trashed when playing remodel"
                )
            gain_card = player.gain_resp(
                card=self,
                valid_cards=[
                    card
                    for card in game.supply.avaliable_cards()
                    if card.cost <= trash_card.cost + 2
                ],
                game=game,
                required=True,
            )
            if not gain_card:
                raise InvalidBotImplementation(
                    "Card must be gained when playing remodel"
                )

        player.trash(trash_card, trash=game.trash)
        player.gain(gain_card, game.supply)


class Mine(Action):
    """
    You may trash a Treasure from your hand. Gain a Treasure to your hand costing up to $3 more than it

    """

    def __init__(
        self,
        name: str = "Mine",
        cost: int = 5,
        type: Tuple[str] = ("Action",),
    ):
        super().__init__(name, cost, type)

    def play(
        self, player: Union[Human, Bot], game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        treasures = [card for card in player.hand.cards if "Treasure" in card.type]

        if not treasures:
            return

        if isinstance(player, Human):

            @validate_input(exceptions=InvalidSingleCardInput)
            def get_trash_card() -> Optional[Card]:

                trash_card = player.single_card_decision(
                    prompt="You may trash a Treasure from your hand: ",
                    valid_cards=treasures,
                )
                if isinstance(trash_card, str):
                    raise InvalidSingleCardInput(f"You can not trash {trash_card}")

                return trash_card

            @validate_input(exceptions=InvalidSingleCardInput)
            def get_gain_card(trash_card: Card) -> Card:

                gain_card = player.single_card_decision(
                    prompt=f"Gain a Treasure card costing up to {trash_card.cost + 3} money to your hand: ",
                    valid_cards=game.supply.avaliable_cards(),
                )

                if not gain_card or isinstance(gain_card, str):
                    raise InvalidSingleCardInput("You must gain a card")
                if "Treasure" not in trash_card.type:
                    raise InvalidSingleCardInput("Card must be a Treasure")
                if gain_card.cost > trash_card.cost + 3:
                    raise InvalidSingleCardInput(
                        f"Card must cost less than {trash_card.cost + 3} money"
                    )
                return gain_card

            trash_card = get_trash_card()
            if not trash_card:
                return
            gain_card = get_gain_card(trash_card)

        elif isinstance(player, Bot):
            trash_card = player.trash_resp(
                card=self,
                valid_cards=treasures,
                game=game,
                required=False,
            )
            if not trash_card:
                return

            gain_card = player.gain_resp(
                card=self,
                valid_cards=[
                    card
                    for card in game.supply.avaliable_cards()
                    if "Treasure" in card.type and card.cost <= trash_card.cost + 3
                ],
                game=game,
                required=True,
            )
            if not gain_card:
                raise InvalidBotImplementation("Card must be gained when playing mine")

        player.trash(trash_card, trash=game.trash)
        player.gain(gain_card, game.supply, destination=player.hand)


class Militia(Action):
    """
    +2 Money

    Each other player discards down to 3 cards in hand

    """

    def __init__(
        self,
        name: str = "Militia",
        cost: int = 4,
        type: Tuple[str] = ("Action", "Attack"),
        money: int = 2,
    ):
        super().__init__(name, cost, type, money=money)

    def play(
        self, player: Union[Human, Bot], game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.state.money += 2

        for opponent in game.players:
            if opponent is not player and opponent.is_attacked(
                player=player, attack_card=self
            ):

                num_discard = len(opponent.hand) - 3
                if num_discard <= 0:
                    return

                @validate_input(exceptions=InvalidMultiCardInput)
                def get_discard_cards() -> List[Card]:
                    cards = opponent.multiple_card_decision(
                        prompt=f"You must discard {num_discard} cards from your hand: ",
                        valid_cards=opponent.hand.cards,
                    )
                    if len(cards) != num_discard:
                        raise InvalidMultiCardInput(
                            f"You must discard {num_discard} cards, you selected {len(cards)}"
                        )
                    return cards

                if isinstance(opponent, Human):
                    discard_cards = get_discard_cards()

                elif isinstance(opponent, Bot):
                    discard_cards = opponent.multiple_discard_resp(
                        card=self,
                        valid_cards=opponent.hand.cards,
                        game=game,
                        num_discard=num_discard,
                        required=True,
                    )
                    if not discard_cards:
                        return

                for card in discard_cards:
                    opponent.discard(target_card=card)


class Sentry(Action):
    """
    +1 card, +1 action

    Look at the top 2 cards of your deck. Trash and/or discard any number of them. Put the rest back on top in any order

    """

    def __init__(
        self,
        name: str = "Sentry",
        cost: int = 5,
        type: Tuple[str] = ("Action",),
        actions: int = 1,
        draw: int = 1,
    ):
        super().__init__(name, cost, type, actions=actions, draw=draw)

    def play(
        self, player: Union[Human, Bot], game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        player.draw()
        player.state.actions += 1

        revealed = AbstractDeck()
        player.draw(num_cards=2, destination=revealed, silent=True)
        logger.info(f"{player} looks at {revealed}")

        if isinstance(player, Human):

            def get_trash_cards() -> Optional[List[Card]]:
                trash_cards = player.multiple_card_decision(
                    prompt="Enter the cards you would like to trash: ",
                    valid_cards=revealed.cards,
                )
                return trash_cards

            def get_discard_cards(
                revealed: AbstractDeck,
            ) -> Optional[List[Card]]:
                if not revealed.cards:
                    return None

                discard_cards = player.multiple_card_decision(
                    prompt="Enter the cards you would like to discard: ",
                    valid_cards=revealed.cards,
                )
                return discard_cards

            trash_cards = get_trash_cards()
            if trash_cards:
                for card in trash_cards:
                    revealed.remove(card)
            discard_cards = get_discard_cards(revealed=revealed)
            if discard_cards:
                for card in discard_cards:
                    revealed.remove(card)
            reorder = False
            if len(revealed.cards) == 2:
                logger.info(
                    f"Current order: {revealed.cards[0]} (Top), {revealed.cards[1]} (Bottom)"
                )
                reorder = player.binary_decision(
                    prompt="Would you like to switch the order of the cards?"
                )

        elif isinstance(player, Bot):
            trash_cards = player.multiple_trash_resp(
                card=self,
                valid_cards=revealed.cards,
                game=game,
                required=False,
            )
            if trash_cards:
                for card in trash_cards:
                    revealed.remove(card)
            discard_cards = player.multiple_discard_resp(
                card=self,
                valid_cards=revealed.cards,
                game=game,
                required=False,
            )
            if discard_cards:
                for card in discard_cards:
                    revealed.remove(card)
            reorder = False
            if len(revealed.cards) == 2:
                reorder = player.binary_resp(game=game, card=self)

        if trash_cards:
            for card in trash_cards:
                game.trash.add(card)
                logger.info(f"{player} trashes {card}")
        if discard_cards:
            for card in discard_cards:
                player.discard_pile.add(card)
                logger.info(f"{player} discards {card}")
        if revealed.cards:
            if reorder:
                for card in revealed.cards:
                    player.deck.add(card)
            else:
                for card in reversed(revealed.cards):
                    player.deck.add(card)
            logger.info(f"{player} topdecks {len(revealed.cards)} cards")


class Library(Action):
    """
    Draw until you have 7 cards in hand, skipping any Action cards you choose to; set those aside, discarding them afterwards

    """

    def __init__(
        self,
        name: str = "Library",
        cost: int = 5,
        type: Tuple[str] = ("Action",),
    ):
        super().__init__(name, cost, type)

    def play(
        self, player: Union[Human, Bot], game: "Game", generic_play: bool = True
    ) -> None:

        logger.info(f"{player} plays {self}")

        if generic_play:
            super().generic_play(player)

        set_aside = AbstractDeck()
        while len(player.hand) < 7:

            if len(player.deck) == 0 and len(player.discard_pile) == 0:
                return

            player.draw(num_cards=1, destination=set_aside)
            drawn_card = set_aside.cards[-1]

            if "Action" in drawn_card.type:
                if isinstance(player, Human):
                    if player.binary_decision(
                        prompt=f"You drew {drawn_card}, would you like to skip it? y/n: "
                    ):
                        pass
                    else:
                        player.hand.add(set_aside.remove(drawn_card))

                    pass
                if isinstance(player, Bot):
                    if player.binary_resp(
                        game=game, card=self, relevant_cards=[drawn_card]
                    ):
                        pass
                    else:
                        player.hand.add(set_aside.remove(drawn_card))

            else:
                player.hand.add(set_aside.remove(drawn_card))

        if set_aside.cards:
            logger.info(f"{player} discards {set_aside}")
            set_aside.move_to(destination=player.discard_pile)


copper = Copper()
silver = Silver()
gold = Gold()
estate = Estate()
duchy = Duchy()
province = Province()
curse = Curse()

artisan = Artisan()
bandit = Bandit()
bureaucrat = Bureaucrat()
cellar = Cellar()
chapel = Chapel()
council_room = CouncilRoom()
festival = Festival()
gardens = Gardens()
harbinger = Harbinger()
laboratory = Laboratory()
library = Library()
market = Market()
merchant = Merchant()
militia = Militia()
mine = Mine()
moat = Moat()
moneylender = Moneylender()
poacher = Poacher()
remodel = Remodel()
sentry = Sentry()
smithy = Smithy()
throne_room = ThroneRoom()
vassal = Vassal()
village = Village()
witch = Witch()
workshop = Workshop()


base_set = [
    artisan,
    bandit,
    bureaucrat,
    cellar,
    chapel,
    council_room,
    festival,
    gardens,
    harbinger,
    laboratory,
    library,
    market,
    merchant,
    militia,
    mine,
    moat,
    moneylender,
    poacher,
    remodel,
    sentry,
    smithy,
    throne_room,
    vassal,
    village,
    witch,
    workshop,
]



import logging
from typing import TYPE_CHECKING, List, Optional, Union

from pyminion.bots.bot import Bot
from pyminion.core import Card
from pyminion.exceptions import InvalidBotImplementation
from pyminion.expansions.base import duchy, estate, silver
from pyminion.players import Player

if TYPE_CHECKING:
    from pyminion.game import Game

logger = logging.getLogger()


class OptimizedBot(Bot):
    """
    Implements opinionated logic for playing and reacting to all cards in the base set.

    The intention is to inherit from this class to make concrete bot implementations.
    If inheriting from this bot, it is possible to change the way that a single card is executed
    by overwriting the card specific method at the bottom of this file.


    """

    def __init__(
        self,
        player_id: str = "bot",
    ):
        super().__init__(player_id=player_id)

    @staticmethod
    def sort_for_discard(cards: List[Card], actions: int) -> List[Card]:
        """
        Sort list of cards from best discard candidate to worst discard candidate.
        First sort cards from lowest cost to highest cost. Then rearrange depending on remaining actions.
        If player has no remaining actions, prioritize discarding victory then action, then treasures.
        If player has remaining actions, prioritize discarding victory then treasure and action equally.

        """

        sorted_cards = sorted(cards, key=lambda card: card.cost)
        victory_cards = [
            card
            for card in sorted_cards
            if "Victory" in card.type or "Curse" in card.type
        ]
        non_victory_cards = [
            card
            for card in sorted_cards
            if "Victory" not in card.type and "Curse" not in card.type
        ]
        treasure_cards = [card for card in non_victory_cards if "Treasure" in card.type]
        action_cards = [
            card for card in non_victory_cards if "Treasure" not in card.type
        ]
        if actions == 0:
            return victory_cards + action_cards + treasure_cards
        else:
            return victory_cards + non_victory_cards

    def determine_trash_cards(
        self, valid_cards: List[Card], player: Player, game: "Game"
    ) -> List[Card]:
        """
        Determine which cards should be trashed:

        Always trash Curse
        Trash Estate if number of provinces in supply >= 5
        Trash Copper if money in deck > 3 (keep enough to buy silver)
        Finally, sort the cards as to prioritize trashing estate over copper


        """
        deck_money = player.get_deck_money()
        trash_cards = []
        for card in valid_cards:
            if card.name == "Curse":
                trash_cards.append(card)
            elif (
                card.name == "Estate"
                and game.supply.pile_length(pile_name="Province") >= 5
            ):
                trash_cards.append(card)
            elif card.name == "Copper" and deck_money > 3:
                trash_cards.append(card)
                deck_money -= 1

        sorted_trash_cards = self.sort_for_discard(cards=trash_cards, actions=1)

        return sorted_trash_cards

    def binary_resp(
        self, card: Card, game: "Game", relevant_cards: Optional[List[Card]] = None
    ) -> bool:
        if card.name == "Moneylender":
            return self.moneylender()
        if card.name == "Vassal":
            return self.vassal(relevant_cards=relevant_cards)
        if card.name == "Sentry":
            return self.sentry(game=game, relevant_cards=relevant_cards, binary=True)
        if card.name == "Library":
            return self.library(relevant_cards=relevant_cards)
        else:
            return True

    def discard_resp(
        self,
        card: Card,
        valid_cards: List[Card],
        game: "Game",
        required: bool = True,
    ) -> Optional[Card]:
        return super().discard_resp(
            card=card, valid_cards=valid_cards, game=game, required=required
        )

    def multiple_discard_resp(
        self,
        card: Card,
        valid_cards: List[Card],
        game: "Game",
        num_discard: Optional[int] = None,
        required: bool = True,
    ) -> Optional[List[Card]]:

        if card.name == "Cellar":
            return self.cellar(valid_cards=valid_cards)
        if card.name == "Poacher":
            return self.poacher(valid_cards=valid_cards, num_discard=num_discard)

        if card.name == "Militia":
            return self.militia(valid_cards=valid_cards, num_discard=num_discard)

        if card.name == "Sentry":
            return self.sentry(game=game, valid_cards=valid_cards, discard=True)

    def gain_resp(
        self,
        card: Card,
        valid_cards: List[Card],
        game: "Game",
        required: bool = True,
    ) -> Optional[Card]:

        if card.name == "Artisan":
            return self.artisan(game=game, gain=True)
        if card.name == "Workshop":
            return self.workshop(game=game)
        if card.name == "Remodel":
            return self.remodel(valid_cards=valid_cards, gain=True)
        if card.name == "Mine":
            return self.mine(valid_cards=valid_cards, gain=True)

    def multiple_gain_resp(
        self,
        card: Card,
        valid_cards: List[Card],
        game: "Game",
        num_gain: Optional[int] = None,
        required: bool = True,
    ) -> Optional[List[Card]]:
        return super().multiple_gain_resp(
            card=card,
            valid_cards=valid_cards,
            game=game,
            num_gain=num_gain,
            required=required,
        )

    def trash_resp(
        self,
        card: Card,
        valid_cards: List[Card],
        game: "Game",
        required: bool = True,
    ) -> Optional[Card]:
        if card.name == "Remodel":
            return self.remodel(valid_cards=valid_cards, trash=True)
        if card.name == "Mine":
            return self.mine(valid_cards=valid_cards, trash=True)

    def multiple_trash_resp(
        self,
        card: Card,
        valid_cards: List[Card],
        game: "Game",
        num_trash: Optional[int] = None,
        required: bool = True,
    ) -> Optional[List[Card]]:

        if card.name == "Chapel":
            return self.chapel(game=game, valid_cards=valid_cards)
        if card.name == "Sentry":
            return self.sentry(game=game, valid_cards=valid_cards, trash=True)

    def topdeck_resp(
        self,
        card: Card,
        valid_cards: List[Card],
        game: "Game",
        required: bool = True,
    ) -> Optional[Card]:

        if card.name == "Artisan":
            return self.artisan(game=game, valid_cards=valid_cards, topdeck=True)
        if card.name == "Harbinger":
            return self.harbinger(valid_cards=valid_cards)

        if card.name == "Bureaucrat":
            return self.bureaucrat(valid_cards=valid_cards)

    def double_play_resp(
        self,
        card: Card,
        valid_cards: List[Card],
        game: "Game",
        required: bool = True,
    ) -> Optional[Card]:
        if card.name == "Throne Room":
            return self.throne_room(valid_cards=valid_cards)

    def is_attacked(self, player: Player, attack_card: Card) -> bool:
        for card in self.hand.cards:
            if card.name == "Moat":
                return False
        return True

    # CARD SPECIFIC IMPLEMENTATIONS
    def moneylender(self) -> bool:
        return True

    def vassal(self, relevant_cards: Optional[List[Card]]) -> bool:
        return True

    def library(self, relevant_cards: Optional[List[Card]]) -> bool:
        if self.state.actions == 0:
            return True
        else:
            return False

    def sentry(
        self,
        game: "Game",
        valid_cards: Optional[List[Card]] = None,
        relevant_cards: Optional[List[Card]] = None,
        trash: bool = False,
        discard: bool = False,
        binary: bool = False,
    ) -> Union[List[Card], bool]:
        if trash:
            if not valid_cards:
                return []
            return self.determine_trash_cards(
                valid_cards=valid_cards, player=self, game=game
            )
        if discard:
            if not valid_cards:
                return []
            return [
                card
                for card in valid_cards
                if card.name == "Copper"
                or "Victory" in card.type
                or "Curse" in card.type
            ]
        if binary:
            return False

        raise InvalidBotImplementation(
            "Either gain, topdeck or binary must be true when playing sentry"
        )

    def cellar(self, valid_cards: List[Card]) -> Optional[List[Card]]:
        return [
            card
            for card in valid_cards
            if card.name == "Copper" or "Victory" in card.type or "Curse" in card.type
        ]

    def poacher(
        self, valid_cards: List[Card], num_discard: Optional[int]
    ) -> List[Card]:
        if not num_discard:
            return []
        discard_order = self.sort_for_discard(
            cards=valid_cards, actions=self.state.actions
        )
        return discard_order[:num_discard]

    def militia(
        self, valid_cards: List[Card], num_discard: Optional[int]
    ) -> List[Card]:
        if not num_discard:
            return []
        discard_order = self.sort_for_discard(
            cards=valid_cards, actions=self.state.actions
        )
        return discard_order[:num_discard]

    def artisan(
        self,
        game: "Game",
        valid_cards: Optional[List[Card]] = None,
        gain: bool = False,
        topdeck: bool = False,
    ) -> Card:
        if topdeck:
            for card in self.hand.cards:
                if "Action" in card.type and self.state.actions == 0:
                    return card
            else:
                return self.hand.cards[-1]
        if gain:
            if game.supply.pile_length(pile_name="Province") < 5:
                return duchy
            else:
                return silver

        raise InvalidBotImplementation(
            "Either gain or topdeck must be true when playing artisian"
        )

    def workshop(self, game: "Game") -> Card:
        if game.supply.pile_length(pile_name="Province") < 3:
            return estate
        else:
            return silver

    def remodel(
        self, valid_cards: List[Card], trash: bool = False, gain: bool = False
    ) -> Card:
        if trash:
            min_price_card = min(valid_cards, key=lambda card: card.cost)
            return min_price_card

        if gain:
            max_price_card = max(valid_cards, key=lambda card: card.cost)
            return max_price_card

        raise InvalidBotImplementation(
            "Either gain or trash must be true when playing remodel"
        )

    def mine(
        self, valid_cards: List[Card], trash: bool = False, gain: bool = False
    ) -> Card:
        if trash:
            min_price_card = min(valid_cards, key=lambda card: card.cost)
            return min_price_card

        if gain:
            max_price_card = max(valid_cards, key=lambda card: card.cost)
            return max_price_card

        raise InvalidBotImplementation(
            "Either gain or trash must be true when playing mine"
        )

    def chapel(self, game: "Game", valid_cards: List[Card]) -> Optional[List[Card]]:
        trash_cards = self.determine_trash_cards(
            valid_cards=valid_cards, player=self, game=game
        )
        while len(trash_cards) > 4:
            trash_cards.pop()
        return trash_cards

    def harbinger(self, valid_cards: List[Card]) -> Optional[Card]:
        # Do not topdeck victory cards
        best_topdeck = [card for card in valid_cards if "Victory" not in card.type]
        if not best_topdeck:
            return None
        # Topdeck highest price card if price > 2
        max_price_card = max(best_topdeck, key=lambda card: card.cost)
        if max_price_card.cost > 2:
            return max_price_card
        else:
            return None

    def bureaucrat(self, valid_cards: List[Card]) -> Card:
        sorted_cards = sorted(valid_cards, key=lambda card: card.cost)
        return sorted_cards[0]

    def throne_room(self, valid_cards: List[Card]) -> Card:
        # Double play most expensive card
        max_price_card = max(valid_cards, key=lambda card: card.cost)
        return max_price_card



import functools
import logging
from collections import Counter
from typing import Callable, List, Optional, Tuple, Type, Union

from pyminion.core import Card
from pyminion.exceptions import (InvalidBinaryInput, InvalidMultiCardInput,
                                 InvalidSingleCardInput)

logger = logging.getLogger()


def validate_input(
    func: Optional[Callable] = None,
    exceptions: Union[Tuple[Type[Exception], ...], Type[Exception]] = (),
):
    """
    Decorator to ensure that a user enters valid input when prompted.
    If input is invalid, the error is printed to the terminal and the user is prompted again

    Accepts a tuple of exceptions to catch

    """
    assert callable(func) or func is None

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            while True:
                try:
                    return func(*args, **kwargs)
                except exceptions as e:
                    logger.error(e)

        return wrapper

    return decorator(func) if callable(func) else decorator


def binary_decision(prompt: str) -> bool:
    """
    Get user response to a binary question.
    Raise exception is input is anything other than 'yes' or 'no'

    """

    decision = input(prompt)
    if decision in ("yes", "y", "Yes", "Y"):
        return True
    elif decision in ("no", "n", "No", "N"):
        return False
    else:
        raise InvalidBinaryInput("Invalid response, valid choices are 'yes' or 'no'")


def single_card_decision(
    prompt: str, valid_cards: List[Card], valid_mixin: str = "placeholder"
) -> Optional[Union[Card, str]]:
    """
    Get user input when given the option to select one card

    valid_cards are a list of cards that a user can choose from. For example,
    when prompting a user to discard a card from their hand, valid cards would
    be the user's hand.

    valid_mixin allows for options other than a card to be selected. For example,
    "all" mixin would be used to autoplay treasures when prompting treasures to play

    Raise exception if user provided selection is not in valid_cards and is not
    a valid_mixin.

    """
    card_input = input(prompt)
    if not card_input:
        return None

    if card_input == valid_mixin:
        return valid_mixin

    for card in valid_cards:
        if card_input.casefold() == card.name.casefold():
            return card

    raise InvalidSingleCardInput(
        f"Invalid input, {card_input} is not a valid selection"
    )


def multiple_card_decision(
    prompt: str, valid_cards: List[Card]
) -> Optional[List[Card]]:
    """
    Get user input when given the option to select multiple cards

    valid_cards are a list of cards that a user can choose from. For example,
    when prompting a user to trash cards from their hand, valid cards would
    be the user's hand.

    Raise exception if user provided selection is not in valid_cards.

    """
    card_input = input(prompt)
    if not card_input:
        return None
    card_strings = [x.strip() for x in card_input.split(",")]
    selected_cards = []
    for card_string in card_strings:
        for card in valid_cards:
            if card_string.casefold() == card.name.casefold():
                selected_cards.append(card)
                break
        else:
            raise InvalidMultiCardInput(
                f"Invalid input, {card_string} is not a valid card"
            )

    selected_count = Counter(selected_cards)
    valid_count = Counter(valid_cards)

    for element in selected_count:
        if selected_count[element] > valid_count[element]:
            raise InvalidMultiCardInput(
                f"Invalid input, attempted to select too many copies of {element}"
            )
    return selected_cards



import logging

from pyminion.bots.bot import Bot
from pyminion.expansions.base import gold, province, silver, smithy
from pyminion.game import Game

logger = logging.getLogger()


class BigMoneySmithy(Bot):
    """
    Big money + smithy

    """

    def __init__(
        self,
        player_id: str = "big_money_smithy",
    ):
        super().__init__(player_id=player_id)

    def action_priority(self, game: Game):
        yield smithy

    def buy_priority(self, game: Game):
        money = self.state.money
        if money >= 8:
            yield province
        if money >= 6:
            yield gold
        if money == 4:
            yield smithy
        if money >= 3:
            yield silver
